/*
 * This script creates a track with parameters according to the user dialog
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2014-12-12
 */

//http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_entity.html

include("../QCTrack.js");
include("../../WidgetFactory.js");

function Create(guiAction) {
    QCTrack.call(this, guiAction);
}

Create.prototype = new EAction();
Create.includeBasePath = includeBasePath;

Create.prototype.beginEvent = function() {
    EAction.prototype.beginEvent.call(this);
    var document = EAction.getDocument();
    var di = EAction.getDocumentInterface();

    //let the user enter the desired parameters
    var dialog = WidgetFactory.createWidget(
                   "scripts/QCTrack/Create",
                   "CreateDialog.ui");
    if (!dialog.exec()) { //the user aborted
        this.terminate();
        return;
    }

    //get and check given parameters
    var widgets = getWidgets(dialog);
    var blockName = widgets["blockName"].text;
    var cost = widgets["cost"].value;
    var displayName = widgets["displayName"].text;
    var displayNameSize = widgets["displayNameSize"].value;
    var trackDistance = widgets["trackDistance"].value;
    var bedSize = widgets["bedSize"].value;
    var swellDistance = widgets["swellDistance"].value;
    var swellWidth = widgets["swellWidth"].value;
    var swellLength = widgets["swellLength"].value;
    var straight = widgets["straight"].value;
    var radius = widgets["radius"].value;
    var angle = widgets["angle"].value;
    if (blockName == "") {
      EAction.handleUserMessage("block name must not be empty");
      dialog.destroy();
      this.terminate();
      return;
    }
    if (document.hasBlock(blockName)) {
      EAction.handleUserMessage("block name already existing");
      dialog.destroy();
      this.terminate();
      return;
    }
    if (trackDistance <= 0) {
      EAction.handleUserMessage("track distance must be positive");
      dialog.destroy();
      this.terminate();
      return;
    }
    if ((straight != 0 && (radius != 0 || angle != 0)) || (straight == 0 && radius == 0 && angle == 0)) {
      EAction.handleUserMessage("either straight or curve (one set has to be greater zero, set other values to zero)");
      dialog.destroy();
      this.terminate();
      return;
    }

    //check required layers
    checkLayerExistence(document, di, "TR_MIDS", "black");
    checkLayerExistence(document, di, "TR_PLAN", "black");
    checkLayerExistence(document, di, "TR_ENDS", "black");
    checkLayerExistence(document, di, "TR_NMBRS", "black");
    checkLayerExistence(document, di, "TR_SWELLS", "orange");
    checkLayerExistence(document, di, "TR_BEDS", "grey");
    if (cost > 0) {
      checkLayerExistence(document, di, "TR_COST", "red");
    }

    //create track with given parameters
    //--prepare
    EAction.handleUserMessage("creating "+blockName);
    var operation;
    var block = new RBlock(document, blockName, new RVector(0,0));
    operation = new RAddObjectOperation(block);
    di.applyOperation(operation);
    di.setCurrentBlock(blockName);
    //--create entities
    if (cost > 0) {
      di.setCurrentColor(new RColor(RColor.ByLayer));
	    di.setCurrentLayer("TR_COST");
      operation = new RAddObjectsOperation();
	    p1 = new RVector(0, 0);
	    text = new RTextEntity(document, new RTextData(
	      p1,                  // position
	      p1,                  // alignment point
	      trackDistance / 4,   // height
	      2.0,                 // text width (ignored for now)
	      RS.VAlignMiddle,     // alignments
	      RS.HAlignLeft,
	      RS.LeftToRight,
	      RS.Exact,
	      1.0,                 // line spacing factor
	      "" + cost,    // the text
	      "Arial",      // font
	      false,        // bold
	      false,        // italic
	      0.0,          // angle
	      false         // simple text without formatting
	    ));
	    operation.addObject(text);
	    di.applyOperation(operation);
    }
    if (straight > 0) { //create straight track
      var p1, p2, p3, p4, line, text;
      di.setCurrentColor(new RColor(RColor.ByBlock));
      di.setCurrentLayer("TR_MIDS");
      operation = new RAddObjectsOperation();
      p1 = new RVector(0, 0);
      p2 = new RVector(straight, 0);
      line = new RLineEntity(document, new RLineData(p1, p2));
      operation.addObject(line);
      di.applyOperation(operation);
      operation = new RAddObjectsOperation();
      di.setCurrentColor(new RColor(RColor.ByBlock));
      di.setCurrentLayer("TR_ENDS");
      p1 = new RVector(0, trackDistance / 2);
      p2 = new RVector(0, -trackDistance / 2);
      p3 = new RVector(straight, -trackDistance / 2);
      p4 = new RVector(straight, trackDistance / 2);
      line = new RLineEntity(document, new RLineData(p1, p2));
      operation.addObject(line);
      line = new RLineEntity(document, new RLineData(p3, p4));
      operation.addObject(line);
      di.applyOperation(operation);
      if (bedSize > 0) {
        di.setCurrentColor(new RColor(RColor.ByLayer));
		    di.setCurrentLayer("TR_BEDS");
		    operation = new RAddObjectsOperation();
		    p1 = new RVector(0, bedSize / 2);
		    p2 = new RVector(0, -bedSize / 2);
		    p3 = new RVector(straight, -bedSize / 2);
		    p4 = new RVector(straight, bedSize / 2);
        addRectangle(document, operation, p1, p2, p3, p4);
		    di.applyOperation(operation);
      }
      if (swellWidth > 0 && swellLength > 0 && swellDistance > 0) {
        di.setCurrentColor(new RColor(RColor.ByLayer));
		    di.setCurrentLayer("TR_SWELLS");
		    operation = new RAddObjectsOperation();
        var i;
        var count = Math.floor((straight - swellWidth) / swellDistance) + 1;
        var baseOffset = (straight - (count - 1) * swellDistance - swellWidth) / 2;
        for (i = 0; i < count; i++) {
          var base = baseOffset + swellDistance * i;
				  p1 = new RVector(base, swellLength / 2);
				  p2 = new RVector(base, -swellLength / 2);
				  p3 = new RVector(base + swellWidth, -swellLength / 2);
				  p4 = new RVector(base + swellWidth, swellLength / 2);
				  addHatch(document, operation, p1, p2, p3, p4);
        }
		    di.applyOperation(operation);
      }
      di.setCurrentColor(new RColor(RColor.ByBlock));
      di.setCurrentLayer("TR_PLAN");
      operation = new RAddObjectsOperation();
      p1 = new RVector(0, trackDistance / 2);
      p2 = new RVector(0, -trackDistance / 2);
      p3 = new RVector(straight, -trackDistance / 2);
      p4 = new RVector(straight, trackDistance / 2);
      addRectangle(document, operation, p1, p2, p3, p4);
      di.applyOperation(operation);
      if (displayNameSize > 0 && displayName != "") {
        di.setCurrentColor(new RColor(RColor.ByLayer));
		    di.setCurrentLayer("TR_NMBRS");
        operation = new RAddObjectsOperation();
		    p1 = new RVector(straight / 2, 0);
		    text = new RTextEntity(document, new RTextData(
		      p1,                  // position
		      p1,                  // alignment point
		      displayNameSize,     // height
		      2.0,                 // text width (ignored for now)
		      RS.VAlignMiddle,     // alignments
		      RS.HAlignCenter,
		      RS.LeftToRight,
		      RS.Exact,
		      1.0,                 // line spacing factor
		      displayName,         // the text
		      "Arial",      // font
		      false,        // bold
		      false,        // italic
		      0.0,          // angle
		      false         // simple text without formatting
		    ));
		    operation.addObject(text);
		    di.applyOperation(operation);
      }
    }
    else { //create curved track
      var center, p1, p2, p3, p4, line, arc, text;
      angle = RMath.deg2rad(angle);
      center = new RVector(0, radius);
      di.setCurrentColor(new RColor(RColor.ByBlock));
      di.setCurrentLayer("TR_MIDS");
      operation = new RAddObjectsOperation();
      arc = new RArcEntity(document, new RArcData(center, radius, 3 * Math.PI / 2, 3 * Math.PI / 2 + angle, false));
      operation.addObject(arc);
      di.applyOperation(operation);
      operation = new RAddObjectsOperation();
      di.setCurrentColor(new RColor(RColor.ByBlock));
      di.setCurrentLayer("TR_ENDS");
      p1 = new RVector(0, trackDistance / 2);
      p2 = new RVector(0, -trackDistance / 2);
      line = new RLineEntity(document, new RLineData(p1, p2));
      operation.addObject(line);
      line = new RLineEntity(document, new RLineData(p2, p1));
      line.rotate(angle, center);
      operation.addObject(line);
      di.applyOperation(operation);
      if (bedSize > 0) {
        di.setCurrentColor(new RColor(RColor.ByLayer));
		    di.setCurrentLayer("TR_BEDS");
        operation = new RAddObjectsOperation();
		    arc = new RArcEntity(document, new RArcData(center, radius - bedSize / 2, 3 * Math.PI / 2, 3 * Math.PI / 2 + angle, false));
		    operation.addObject(arc);
		    arc = new RArcEntity(document, new RArcData(center, radius + bedSize / 2, 3 * Math.PI / 2, 3 * Math.PI / 2 + angle, false));
		    operation.addObject(arc);
		    p1 = new RVector(0, bedSize / 2);
		    p2 = new RVector(0, -bedSize / 2);
		    line = new RLineEntity(document, new RLineData(p1, p2));
		    operation.addObject(line);
		    line = new RLineEntity(document, new RLineData(p2, p1));
		    line.rotate(angle, center);
		    operation.addObject(line);
		    di.applyOperation(operation);
      }
      if (swellWidth > 0 && swellLength > 0 && swellDistance > 0) {
        di.setCurrentColor(new RColor(RColor.ByLayer));
		    di.setCurrentLayer("TR_SWELLS");
		    operation = new RAddObjectsOperation();
        var i;
        var arclen = radius * angle;
        var count = Math.floor((arclen - swellWidth) / swellDistance) + 1;
        var baseAngle = (angle / count * (swellDistance - swellWidth) / swellDistance) / 2;
        var angleDiff = angle / count;
        for (i = 0; i < count; i++) {
          var curAngle = baseAngle + angleDiff * i;
		      p1 = new RVector(0, swellLength / 2);
		      p2 = new RVector(0, -swellLength / 2);
		      p3 = new RVector(swellWidth, -swellLength / 2);
		      p4 = new RVector(swellWidth, swellLength / 2);
          p1.rotate(curAngle, center);
          p2.rotate(curAngle, center);
          p3.rotate(curAngle, center);
          p4.rotate(curAngle, center);
				  addHatch(document, operation, p1, p2, p3, p4);
        }
		    di.applyOperation(operation);
      }
      di.setCurrentColor(new RColor(RColor.ByBlock));
      di.setCurrentLayer("TR_PLAN");
      operation = new RAddObjectsOperation();
      arc = new RArcEntity(document, new RArcData(center, radius - trackDistance / 2, 3 * Math.PI / 2, 3 * Math.PI / 2 + angle, false));
      operation.addObject(arc);
      arc = new RArcEntity(document, new RArcData(center, radius + trackDistance / 2, 3 * Math.PI / 2, 3 * Math.PI / 2 + angle, false));
      operation.addObject(arc);
      p1 = new RVector(0, trackDistance / 2);
      p2 = new RVector(0, -trackDistance / 2);
      line = new RLineEntity(document, new RLineData(p1, p2));
      operation.addObject(line);
      line = new RLineEntity(document, new RLineData(p2, p1));
      line.rotate(angle, center);
      operation.addObject(line);
      di.applyOperation(operation);
      if (displayNameSize > 0 && displayName != "") {
        di.setCurrentColor(new RColor(RColor.ByLayer));
		    di.setCurrentLayer("TR_NMBRS");
        operation = new RAddObjectsOperation();
		    p1 = new RVector(0, 0);
		    text = new RTextEntity(document, new RTextData(
		      p1,                  // position
		      p1,                  // alignment point
		      displayNameSize,     // height
		      2.0,                 // text width (ignored for now)
		      RS.VAlignMiddle,     // alignments
		      RS.HAlignCenter,
		      RS.LeftToRight,
		      RS.Exact,
		      1.0,                 // line spacing factor
		      displayName,         // the text
		      "Arial",      // font
		      false,        // bold
		      false,        // italic
		      0.0,          // angle
		      false         // simple text without formatting
		    ));
        text.rotate(angle / 2, center);
		    operation.addObject(text);
		    di.applyOperation(operation);
      }
    }
    //--cleanup
    di.setCurrentBlock(RBlock.modelSpaceName);
    EAction.handleUserMessage("done, created "+blockName);
    dialog.destroy();
    this.terminate();
};

function checkLayerExistence(document, di, layerName, colorName) {
  if (!document.hasLayer(layerName)) {
    var color = new RColor(colorName);
    var linetypeContinuousId = document.getLinetypeId("continuous");
    var myLayer = new RLayer(document, layerName, false, false, color, linetypeContinuousId, RLineweight.Weight025);
    var opAddLayer = new RAddObjectOperation(myLayer);
    di.applyOperation(opAddLayer);
  }
}

function addRectangle(document, operation, p1, p2, p3, p4) {
  var line;
  line = new RLineEntity(document, new RLineData(p1, p2));
  operation.addObject(line);
  line = new RLineEntity(document, new RLineData(p2, p3));
  operation.addObject(line);
  line = new RLineEntity(document, new RLineData(p3, p4));
  operation.addObject(line);
  line = new RLineEntity(document, new RLineData(p4, p1));
  operation.addObject(line);
}

function addHatch(document, operation, p1, p2, p3, p4) {
  var data = new RHatchData(true, 0, 0, "");
  data.newLoop();
  data.addBoundary(new RLine(p1, p2));
  data.addBoundary(new RLine(p2, p3));
  data.addBoundary(new RLine(p3, p4));
  data.addBoundary(new RLine(p4, p1));
  var hatch = new RHatchEntity(document, data);
  operation.addObject(hatch);
}

Create.init = function(basePath) {
    var action = new RGuiAction("&Create", RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setScriptFile(basePath + "/Create.js");
    action.setSortOrder(100);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
};
