/*
 * This script moves + rotates the current selection so that
 * the track end selected by user connects to another track. The
 * original track is removed.
 * Attention! This script is highly re-used by DupMove, don't
 * change it without checking DupMove!
 *
 * Copyright smf, some ideas taken from the QCad example scripts
 * last change: 2014-12-13
 */

include("../QCTrack.js");
include("../../Modify/Transform.js");

function Move(guiAction) {
    Transform.call(this, guiAction);

    this.lineSource = undefined;
    this.lineDest = undefined;
}

Move.prototype = new Transform();

Move.includeBasePath = includeBasePath;

Move.State = {
    SettingLineSource : 0,
    SettingLineDest : 1
};

Move.prototype.beginEvent = function() {
    Transform.prototype.beginEvent.call(this);

    if (!this.verifySelection()) {
        return;
    }

    this.continueAfterMove = false;
    this.copies = 0; //delete original
    this.useCurrentAttributes = true; //use current attributes
    this.forceNewObject = false;
    this.setState(Move.State.SettingLineSource);
};

Move.prototype.setState = function(state) {
    Transform.prototype.setState.call(this, state);

    this.getDocumentInterface().setClickMode(RAction.PickEntity);
    this.setCrosshairCursor();

    var appWin = RMainWindowQt.getMainWindow();
    var info;
    switch (this.state) {
    case Move.State.SettingLineSource:
      this.lineSource = undefined;
      this.lineDest = undefined;
      info = qsTr("Select end of track to move");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trCancel);
      break;
    case Move.State.SettingLineDest:
      this.lineDest = undefined;
      info = qsTr("Select destination track end");
      appWin.setCommandPrompt(info);
      appWin.setLeftMouseTip(info);
      appWin.setRightMouseTip(EAction.trBack);
      break;
    }
};

Move.prototype.escapeEvent = function() {
    switch (this.state) {
    case Move.State.SettingLineSource:
      Transform.prototype.escapeEvent.call(this);
      break;
    case Move.State.SettingLineDest:
      this.setState(Move.State.SettingLineSource);
      break;
    }
};

Move.prototype.pickEntity = function(event, preview) {
    var di = this.getDocumentInterface();
    var document = this.getDocument();
    var entityId = event.getEntityId();
    var entity = document.queryEntity(entityId);
    var pos = event.getModelPosition();
    var op;

    switch (this.state) {
    case Move.State.SettingLineSource:
      if (isNull(entity)) {
        this.lineSource = undefined;
        this.lineDest = undefined;
        this.lineTarget = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineSource = getClosestTrackEnd(document, entity, pos, true);
      if (this.lineSource == undefined) {
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      if (preview) {
        this.updatePreview();
      }
      else {
        this.setState(Move.State.SettingLineDest);
      }
      break;
    case Move.State.SettingLineDest:
      if (isNull(entity)) {
        this.lineDest = undefined;
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      this.lineDest = getClosestTrackEnd(document, entity, pos, false);
      if (this.lineDest == undefined) {
        if (preview) {
            this.updatePreview();
        }
        break;
      }
      if (preview) {
        this.updatePreview();
      }
      else {
        var oldSelection = document.querySelectedEntities();
        op = this.getOperation(false);
        if (op == undefined) break;
        di.applyOperation(op);
        if (this.continueAfterMove) { //this is needed for DupMove
          di.selectEntities(oldSelection);
          this.lineDest = undefined;
          this.updatePreview();
        }
        else { //this is for normale move
		      this.setState(-1);
		      this.terminate();
        }
      }
      break;
    }
};

function getClosestTrackEnd(document, block, pos, onlySelected) {
    if (!isBlockReferenceEntity(block)) {
      return undefined;
    }
    var data = block.getData();
    var subIds = document.queryBlockEntities(data.getReferencedBlockId());
    var k, bestDist = -1, bestEnd = undefined;
    for (k = 0; k < subIds.length; k++) {
      var line = data.queryEntity(subIds[k]);
      if (!isLineEntity(line) || line.getLayerName() != "TR_ENDS" || (onlySelected && !line.isSelected())) {
        continue;
      }
      var curDist = pos.getDistanceTo(line.getClosestPointOnEntity(pos));
      if (bestDist == -1 || curDist < bestDist) {
        bestDist = curDist;
        bestEnd = line;
      }
    }
    return bestEnd;
}

Move.prototype.getOperation = function(preview) {
    if (!isLineEntity(this.lineSource) || !isLineEntity(this.lineDest)) {
      return undefined;
    }

    //see http://www.ribbonsoft.com/doc/qcad/3.0/developer/class_r_vector.html
    var moveX, moveY;
    moveX = this.lineDest.getEndPoint().getX() - this.lineSource.getStartPoint().getX();
    moveY = this.lineDest.getEndPoint().getY() - this.lineSource.getStartPoint().getY();
    this.moveBy = new RVector(moveX, moveY);
    this.angle = RMath.getNormalizedAngle(this.lineDest.getAngle() - this.lineSource.getAngle() + Math.PI);
    if (!preview) {
      EAction.handleUserMessage("move by " + moveX + " / " + moveY + ", rotate by " + (this.angle * 180 / Math.PI) + "°");
    }
    return Transform.prototype.getOperation.call(this, preview);
};

Move.prototype.transform = function(entity, k, op) {
    entity.move(this.moveBy);
    entity.rotate(this.angle, this.lineDest.getEndPoint());
    op.addObject(entity, this.useCurrentAttributes, this.forceNewObject);
};

Move.prototype.getAuxPreview = function() {
    var ret = undefined;
    if (isLineEntity(this.lineSource)) {
      ret = new Array();
      var mid = getPointOnLine(this.lineSource, 0.5);
      ret.push(new RCircle(mid, this.lineSource.getLength() * 1.5));
      var top = getPointOnLine(this.lineSource, 2);
      var bottom = getPointOnLine(this.lineSource, -1);
      var out = getPointOnLine(this.lineSource, 2).rotate(Math.PI / 2, mid);
      ret.push(new RLine(top, bottom));
      ret.push(new RLine(bottom, out));
      ret.push(new RLine(out, top));
    }
    return ret;
};

function getPointOnLine(lineEntity, alpha) {
    var start = lineEntity.getStartPoint();
    var end = lineEntity.getEndPoint();
    return new RVector(start.getX() * alpha + end.getX() * (1 - alpha), start.getY() * alpha + end.getY() * (1 - alpha));
}

Move.init = function(basePath) {
    var action = new RGuiAction(qsTr("&Move"), RMainWindowQt.getMainWindow());
    action.setRequiresDocument(true);
    action.setRequiresSelection(true);
    action.setScriptFile(basePath + "/Move.js");
    action.setIcon(basePath + "/Move.svg");
    action.setDefaultShortcut(new QKeySequence("q,m"));
    action.setSortOrder(501);
    EAction.addGuiActionTo(action, QCTrack, true, false, false);
    var appWin = EAction.getMainWindow();
    appWin.addSelectionListener(action);
};
